/*
 * BASSORT - Program to sort a CSG or ICOMMS file by date and time.
 * Note: For CSG, the date is column 9 and the time is column 12.
 *       For ICOMMS, the date is column 9 and the time is column 10.
 *       For OCVS, the date and time are column 9.
 *
 * Arguments: a list of files to sort followed by the path that holds all
 * the files. The filenames are free-form except that the first name must
 * contain "CSG" or "ICOMMS".
 *
 * The format of the date is assumed to be mm/dd/yyyy, where month and day
 * are 1 or 2 digits and year is 4 digits. The delimiter can be any non-digit.
 *
 * The format of the time is assumed to be hh:mm:ss, where hours, minutes,
 * and seconds are 1 or 2 digits (possibly beginning with a space).
 * The delimiter can be any non-digit.
 *
 * The input file is passed on the command line.
 * The output file has the same name as the input file (output replaces input).
 *
 * If the input file is seen to be already sorted, it is simply renamed back
 * to its original name.
 */
#define DBNTWIN32
#define _CRT_SECURE_NO_WARNINGS

#define REPAIRBADDATES 1

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#define LINELEN 1000
#define TEMPSPACESIZE 500000    // size of each block of temporary freespace

#define CSG_STATEDATE    8      // 0-based
#define CSG_STATETIME   11

#define ICM_STATEDATE    8
#define ICM_STATETIME    9
#define ICM_RECEIPTDATE 11

#define OCV_STATEDATE    8
#define OCV_STATETIME    8

#define FUDGEDATE "01/01/2000"

// Binary tree data.

typedef struct Bintreenode {
    long balance;               // tree balance: -1, 0, 1 (right)
    struct Bintreenode *llink;  // pointer to left-hand child
    struct Bintreenode *rlink;  // pointer to right-hand child
    struct Bintreenode *elink;  // pointer to other entries with same key
    time_t key;                 // key value
    char *dataptr;              // pointer to the data
} bintreenode;

bintreenode *treeroot;          // root of the binary tree
bintreenode rootnode;

long nodelevel = -1;            // values to support firstnode, nextnode
long nodestate;
bintreenode *thisnode;
bintreenode *samekeynode;
bintreenode *nodelevels[50];
bintreenode *curnode;

// Space allocation global variables.

char *mytempspaceptr, *mytempspacelist, *mytempspacefree;
long mytempspacesize = 0;       // size of remainder of current block

// General global variables.

struct tm tmptr;                // Date/time of current record
char ArchiveFileName[200];
char ActualFileName[200];
char Header[LINELEN + 1];

FILE *ifh, *ofh;
char line[LINELEN + 1], *LinePtr;
time_t keydate;
time_t prevdate;
short Skip1, Skip2;             // fields to skip to get to date, time
short InputFileIsSorted;

// Local prototypes.

short IsHeader( void);
short LineIsOK( void);
time_t GetDateAndTime( void);
void SkipColumns( short count);
void RepairDate( void);

void treeinit( void);
bintreenode *treestore( long key, char *dataptr);
long treewalkinit( void);
bintreenode *treewalk( void);
char *mytempmalloc( long size);

void main( long argc, char *argv[])
{
    char *tmpstg, *FolderPtr, *AcctSys;
    short argcix, NormalFile;
    long LineNo;

#if 0 //@@ DEBUG
    short ii;
#endif

#if 0 //@@ DEBUG
    for (ii = 0; ii < argc; ++ii) printf( "%s\n", argv[ii]);
    fgets( line, LINELEN, stdin);  // DEBUG - pause before exiting
#endif

    if (argc < 3) {
        printf( "Usage: BASSort <input file(s)> <folder name>\n\n");
        //@@ goto ExitAll;
        exit( 0);
    }

    treeinit();
    LineNo = 0;
    InputFileIsSorted = 1;

    // Figure out the billing system. Note that if there are multiple files,
    // the last argument is the directory they are all in.
    // If neither "CSG" nor "ICOMMS" appears in the filename, use "CSG".

    if (strstr( argv[1], "ICOMMS")) {
        AcctSys = "ICOMMS";
        Skip1 = ICM_STATEDATE;
        Skip2 = ICM_STATETIME - ICM_STATEDATE;
    }
    else if (strstr( argv[1], "OCV")) {
        AcctSys = "OCV";
        Skip1 = OCV_STATEDATE;
        Skip2 = OCV_STATETIME - OCV_STATEDATE;
    }
    else {
        AcctSys = "CSG";
        Skip1 = CSG_STATEDATE;
        Skip2 = CSG_STATETIME - CSG_STATEDATE;
    }

    // If the input file is a normal inbound file, set up its rename name.
    // (Otherwise it's one or more SORT-* files).

    FolderPtr = argv[--argc];
    if ((NormalFile = strncmp( argv[1], "SORT-", 5)))
        sprintf( ArchiveFileName, "%s\\%s.org", FolderPtr, argv[1]);


    // Loop through all the given input files.

    for (argcix = 1; argcix < argc; ++argcix) {
        sprintf( ActualFileName, "%s\\%s", FolderPtr, argv[argcix]);

        // Open the input file.

        if ((ifh = fopen( ActualFileName, "r")) == NULL) {
            printf( "The file ""%s"" was not found\n", ActualFileName);
            //@@ goto ExitAll;
            exit( 1);
        }

        // If the first line is a header, hold on to it. Otherwise process it.

        fgets( line, LINELEN, ifh);
        if (!IsHeader()) goto NoHeader;
        if (!*Header) strcpy( Header, line);

        // Read each line of the input file and build the binary tree.

        while (fgets( line, LINELEN, ifh)) {
NoHeader:
            if (!LineIsOK()) {
                InputFileIsSorted = 0;
                goto SkipRecord;
            }

            if ((keydate = GetDateAndTime()) == -1) {
                RepairDate();
                goto NoHeader;
            }

            if (prevdate > keydate) InputFileIsSorted = 0;
            prevdate = keydate;

            tmpstg = mytempmalloc( (LONG) (strlen( line) + 1));
            strcpy( tmpstg, line);

            treestore( ((long) keydate), tmpstg);
SkipRecord:
            if (!(++LineNo % 100000)) printf( "Read %d records\n", LineNo);
        }
        fclose( ifh);
    }

    // If it's a normal inbound file and the records are already sorted,
    // there's nothing to do. Just exit.
    // Otherwise move the input file aside and write out the sorted records.

    if (!InputFileIsSorted || !NormalFile) {
        // If one input file, rename the it so it's out of the way.
        // If more than one, build a plausible output file name.

        if (NormalFile) rename( ActualFileName, ArchiveFileName);
        else
            sprintf( ActualFileName, "%s\\SINBOUND-%4d%02d%02d%s.csv",
                FolderPtr,
                tmptr.tm_year + 1900, tmptr.tm_mon + 1, tmptr.tm_mday,
                AcctSys);

        // Open the output file. If there's a header line, output it.

        if ((ofh = fopen( ActualFileName, "w")) == NULL) {
            printf( "Could not create ""%s""\n", ActualFileName);
            //@@ goto ExitAll;
            exit( 2);
        }

        if (*Header) fprintf( ofh, "%s", Header);

        // Now walk the tree and write out the sorted records.

        LineNo = 0;
        if (treewalkinit()) {
            while (curnode = treewalk()) {
                fprintf( ofh, "%s", curnode->dataptr);
                if (!(++LineNo % 100000)) printf( "Output %d records\n", LineNo);
            }
        }

        fclose( ofh);

        // Everything completed successfully. Delete the original file.

        if (NormalFile) DeleteFile( ArchiveFileName);
    }
    else printf( "The file is already sorted\n");

    // Delete all the input files, if they're SORT-* files.

    if (!NormalFile) {
        for (argcix = 1; argcix < argc; ++argcix) {
            sprintf( ActualFileName, "%s\\%s", FolderPtr, argv[argcix]);
            DeleteFile( ActualFileName);
        }
    }

    printf( "DONE\n");

//@@ ExitAll:
//@@     fgets( line, LINELEN, stdin);  // DEBUG - pause before exiting
    exit( 0);
}


/*
 * ISHEADER - Look at the date and time fields of the current record.
 * This is a header line iff they both contain no digits.
 */
short IsHeader( void)
{
    char *sptr, chr;

    // Examine the date field in the record.

    LinePtr = line;
    SkipColumns( Skip1);
    sptr = LinePtr;
    while ((chr = *sptr++) && (chr != ','))
        if ((chr >= '0') && (chr <= '9')) return 0;

    // No digits in the date field. Now check the time field.

    if (Skip2) {
        SkipColumns( Skip2);

        while ((chr = *sptr++) && (chr != ','))
            if ((chr >= '0') && (chr <= '9')) return 0;
    }

    // No digits anywhere. It's a header.

    return 1;
}


/*
 * LINEISOK - If there are any illegal characters in the line, return 0, else 1.
 */
short LineIsOK( void)
{
    char *sptr, chr;

    sptr = line;
    while ((chr = *sptr++) && (chr != '\n'))
        if ((chr < ' ') || (chr > 0177)) return 0;

    return 1;
}


/*
 * GETDATEANDTIME - Convert the date and time fields from the current record
 * into a time_t, which is returned.
 */
time_t GetDateAndTime( void)
{
    char *sptr, chr;
    long values[3], lval;
    short vix;

    // Find the date field in the record.

    LinePtr = line;
    SkipColumns( Skip1);

    // Put the three values (separated by some non-digit delimiter)
    // into the values array.

    vix = (short) (values[0] = values[1] = values[2] = 0);
    sptr = LinePtr;

    if (*sptr == '"') ++sptr;
    while (chr = *sptr++) {
        if ((chr >= '0') && (chr <= '9'))
            values[vix] = (values[vix] * 10) + chr - '0';
        else if (++vix > 2) break;
    }

    // If the year comes first, assume it's yyyy/mm/dd, and reorder.

    if (values[0] > 1900) {
        lval = values[0];
        values[0] = values[1];
        values[1] = values[2];
        values[2] = lval;
    }

    tmptr.tm_year = (short) values[2] - 1900;
    tmptr.tm_mon = (short) values[0] - 1;
    tmptr.tm_mday = (short) values[1];

    // Find the time field in the record.

    if (Skip2) {
        SkipColumns( Skip2);
        sptr = LinePtr;
    }

    // Put the three values (separated by some non-digit delimiter)
    // into the values array.

    vix = (short) (values[0] = values[2] = 0);
    values[1] = -1;
    while (chr = *sptr++) {
        if ((chr != ' ') && (chr != '"')) {
            if ((chr >= '0') && (chr <= '9'))
                values[vix] = (values[vix] * 10) + chr - '0';
            else if ((++vix > 2) || (chr == ',')) break;
            if (values[vix] == -1) values[vix] = 0;
        }
    }

    if ((chr == 'P') && (values[0] != 12)) values[0] += 12;
    if ((chr == 'A') && (values[0] == 12)) values[0] -= 12;

    // If there were no delimiters in the time, separate the single value.
    // In this case the time is assumed to be in the form hhmmss.

    if (values[1] == -1) {
        values[2] = values[0] % 100;
        values[1] = (values[0] / 100) % 100;
        values[0] /= 10000;
    }

    tmptr.tm_hour = (short) values[0];
    tmptr.tm_min = (short) values[1];
    tmptr.tm_sec = (short) values[2];
    tmptr.tm_isdst = 0;

    // Convert it to a time_t and return it.

    return mktime( &tmptr);
}


/*
 * SKIPCOLUMNS - Update the pointer LinePtr to skip past the given number
 * of columns. Columns are separated by a comma, and may be in double quotes.
 */
void SkipColumns( short count)
{
    char chr;

    while (chr = *LinePtr++) {
        if (chr == '"') {
            while ((chr = *LinePtr++) && (chr != '"'));
            if (!chr) return;
            continue;
        }

        if ((chr == ',') && !--count) return;
    }
}

/*****************************/
/* FORMAT REPAIR SUBROUTINES */
/*****************************/

/*
 * REPAIRDATE - If the date of state is too early, in the wrong format, or
 * missing, replace it with the box's receipt date. The the latter is also
 * missing, replace it with FUDGEDATE.
 */
void RepairDate( void)
{
    char *dptr, *edptr, *rptr, chr;
    char RDate[15], *rdptr;
    short len, rdlen;

    //  Find the date string
    //  Find the end of the date string
    //  Paste the picked string

    LinePtr = line;
    SkipColumns( ICM_STATEDATE);
    dptr = LinePtr;

    SkipColumns( 1);
    edptr = LinePtr - 1;

    //  Find the receipt string, pick it up

    SkipColumns( ICM_RECEIPTDATE - ICM_STATEDATE - 1);
    rptr = LinePtr;

    //  If receipt string is empty, pick up FUDGEDATE

    if (*rptr != ',') {
        rdptr = RDate;
        while ((chr = *rptr++) != ',') *rdptr++ = chr;
        *rdptr = '\0';
    }
    else strcpy( RDate, FUDGEDATE);
    rdlen = (short) strlen( RDate);

    //  If the date string is not the same length,
    //      open or close the string

    if (len = (rdlen - (short) (edptr - dptr))) {
        if (len > 0) {                  // open up a hole to hold the date
            memcpy( dptr + len, dptr, strlen( dptr) + 1);
        }
        else {                          // close up the hole
            strcpy( dptr, dptr - len);
        }
    }

    // Copy the reasonable date into the hole.

    memcpy( dptr, RDate, rdlen);
}


/***************************/
/* BINARY TREE SUBROUTINES */
/***************************/

/*
 * TREEINIT - Initialize the binary tree.
 */
void treeinit( void)
{
    treeroot = &rootnode;
    rootnode.llink = rootnode.rlink = rootnode.elink = NULL;
    rootnode.balance = 0;
}


/*
 * TREESTORE - Store the given value in the tree.
 * Returns NULL on malloc failure, else the pointer to the entry.
 * This is right out of Knuth vol 3, pages 451-456.
 */
bintreenode *treestore( long key, char *dataptr)
{
    bintreenode *p, *q, *r, *s, *t;
    long alpha;

    // A1 (initialize)
    // If the tree is empty, insert the first entry.

    if (!(s = p = treeroot->rlink)) {
        if (!(treeroot->rlink = q =
            (bintreenode *) mytempmalloc( sizeof( bintreenode))))
                return NULL;
        q->key = key;
        q->dataptr = dataptr;
        q->llink = q->rlink = q->elink = NULL;
        q->balance = 0;
        return q;
    }

    // A2 (compare)

    t = treeroot;
    while (1) {
        // Records with the same key are chained, to save tree width.

        if (key == p->key) {
            if (!(q = (bintreenode *) mytempmalloc( sizeof( bintreenode))))
                return NULL;
            q->key = key;
            q->dataptr = dataptr;
            q->elink = p->elink;
            p->elink = q;
            return q;
        }

        // A3 (move left)

        if (key < p->key) {
            if (!(q = p->llink)) {
                if (!(q = (bintreenode *) mytempmalloc( sizeof( bintreenode))))
                    return NULL;
                p->llink = q;
                break;
            }
        }

        // A4 (move right)

        else {
            if (!(q = p->rlink)) {
                if (!(q = (bintreenode *) mytempmalloc( sizeof( bintreenode))))
                    return NULL;
                p->rlink = q;
                break;
            }
        }

        if (q->balance) {
            t = p;
            s = q;
        }
        p = q;
        continue;
    }


    // A5 (insert)

    q->key = key;
    q->dataptr = dataptr;
    q->llink = q->rlink = q->elink = NULL;
    q->balance = 0;

    // A6 (adjust balance factors)

    if (key < s->key) p = s->llink;
    else              p = s->rlink;
    r = p;

    while (p != q) {
        if (key < p->key) {
            p->balance = -1;
            p = p->llink;
        }
        else {
            p->balance = 1;
            p = p->rlink;
        }
    }

    // A7 (balancing act)

    alpha = (key < s->key) ? -1 : 1;
    if (!s->balance) {
        s->balance = alpha;
        return q;
    }

    if (s->balance == -alpha) {
        s->balance = 0;
        return q;
    }

    // A8 (single rotation)

    if (r->balance == alpha) {
        p = r;

        if (alpha == 1) {
            s->rlink = r->llink;
            r->llink = s;
        }
        else {
            s->llink = r->rlink;
            r->rlink = s;
        }
        s->balance = r->balance = 0;
    }

    // A9 (double rotation)

    else {
        if (alpha == 1) {
            p = r->llink;

            r->llink = p->rlink;
            p->rlink = r;
            s->rlink = p->llink;
            p->llink = s;
        }
        else {
            p = r->rlink;

            r->rlink = p->llink;
            p->llink = r;
            s->llink = p->rlink;
            p->rlink = s;
        }
        if (!p->balance) s->balance = r->balance = 0;
        else if (p->balance == alpha) {
            s->balance = -alpha;
            r->balance = 0;
        }
        else {
            s->balance = 0;
            r->balance = alpha;
        }
        p->balance = 0;
    }

    // A10 (finishing touch)

    if (s == t->rlink) t->rlink = p;
    else t->llink = p;
    return q;
}


/*
 * TREEWALKINIT - Set up to walk the binary tree.
 * Returns 1 if there's anything in the tree, 0 if it's empty.
 */
long treewalkinit( void)
{
    if (!treeroot) return 0;

    nodelevel = nodestate = 0;
    samekeynode = thisnode = treeroot->rlink;
    return (thisnode != NULL);
}


/*
 * TREEWALK - Returns the next node of the tree that's currently being
 * traversed. The caller MUST have done a treewalkinit some time in the past.
 * Returns NULL when the tree is exhausted.
 */
bintreenode *treewalk( void)
{
    if (nodestate) {
        if (samekeynode->elink) {
            samekeynode = samekeynode->elink;
            return samekeynode;
        }

        while (!(thisnode = thisnode->rlink)) {
            if (!nodelevel--) return NULL;
            samekeynode = thisnode = nodelevels[nodelevel];
            return thisnode;
        }
    }

    while (nodelevel >= 0) {        // go as far as possible down left links
        if (thisnode->llink) {
            nodelevels[nodelevel++] = thisnode;
            thisnode = thisnode->llink;
            continue;
        }
        nodestate = 1;
        samekeynode = thisnode;
        return thisnode;
    }
    nodestate = 0;
    return NULL;
}


/*
 * MYTEMPMALLOC - Allocate temporary storage en masse.
 */
char *mytempmalloc( long size)
{
    char *returnarg, **ppptr;

    // If there is no room left in the current block, get another.
    // Use a block on the free list, if there is one, else allocate a new one.

    if (mytempspacesize < size) {
        if (mytempspacefree) {
            mytempspaceptr = mytempspacefree;
            ppptr = (char **) mytempspaceptr;
            mytempspacefree = *ppptr;
        }
        else {
            mytempspaceptr = (char *) malloc( TEMPSPACESIZE);
            ppptr = (char **) mytempspaceptr;
        }
        mytempspacesize = TEMPSPACESIZE - sizeof( char *);

        *ppptr = mytempspacelist;
        mytempspacelist = mytempspaceptr;
        mytempspaceptr += sizeof( char *);
    }

    // Chop off the requested space and return a pointer to it.

    returnarg = mytempspaceptr;
    mytempspaceptr += size;
    mytempspacesize -= size;
    return returnarg;
}
